#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 15 09:59:25 2022

@author: nurmelaj
"""


from datetime import datetime, timedelta
from pygeodesy.ellipsoidalVincenty import LatLon
import os
import sys
import time
import xarray as xr
import cartopy.crs as ccrs
import cartopy.io.img_tiles as cimgt
import matplotlib.pyplot as plt
from netCDF4 import Dataset
import numpy as np
import pandas as pd
from scipy.optimize import least_squares


import warnings
warnings.filterwarnings('ignore', category=RuntimeWarning)

cwd = os.path.abspath(os.path.join('..'))
ddeq_path = os.path.join(cwd, 'ddeq/src/')
'''
if ddeq_path not in sys.path:
    sys.path.append(ddeq_path)
'''
import ddeq
from ddeq.mcmc_tools import AM_MC
from ddeq.divergence_model import model_residual,peak_model,local_divergence_map,global_divergence_map,around_source

def estimate_emission(sources, data_path, wind_path,
                      lon_km=50, lat_km=50, grid_reso=5, UTC='11:00:00',
                      sats=['0000','0805','1610'], wind_product='ERA5',
                      co2_noise_scenario='low', co2_cloud_threshold=0.01,
                      no2_noise_scenario='low', no2_cloud_threshold=0.30,
                      start_date='2015-01-01', end_date='2015-12-24',
                      AM_options=None,trace_gases=['CO2']):
    """ 
    Estimates the NOx and CO2 emissions for one source over the chosen time series
    by computing the divergence map using the SMARTCARB data with the chosen satellites
    and fitting a peak fitting model.
    """
    # Validate supported trace gases
    for gas in trace_gases:
        if gas not in ['CO2','NO2']:
            msg = f'Trace gas {gas} not supported'
            sys.exit(msg)
        source_names = sources['source'].data
    #Process start time
    estimation_start = time.process_time()
    #Dates
    start_date = datetime.strptime(start_date,'%Y-%m-%d') if isinstance(start_date, str) else start_date
    end_date = datetime.strptime(end_date,'%Y-%m-%d') if isinstance(end_date, str) else end_date
    print('Reading wind data...')
    wind = read_wind_data(wind_path, start_date, end_date, dataset=wind_product)
    print('...done')
    
    #Dict to temporalily hold the results
    out = {gas: {source_name: {} for source_name in source_names} for gas in trace_gases}
    maps_xr = {}
    for gas in trace_gases:
        emissions = pd.read_csv(f'{ddeq_path}/ddeq/data/ts_{gas}_emissions.dat',delim_whitespace=True,parse_dates=[['date','UTC']])
        emissions.set_index('date_UTC',inplace=True)
        for source_name in source_names:
            source_xr = sources.loc[{'source':source_name}]
            source_coord = (float(source_xr['lon_o'].data), float(source_xr['lat_o'].data))
            # Regular km grid around the source
            longrid, latgrid, longrid_km, latgrid_km = grids(source_coord,lon_km,lat_km,grid_reso)
            latmin, latmax = np.min(latgrid), np.max(latgrid)
            lonmin, lonmax = np.min(longrid), np.max(longrid)
            # Focus data to the area of interest in order to make the gridding lighter
            if wind_product == 'ERA5':
                enlarge = 0.5
                # Limit data to make the gridding lighter
                cols = np.where((lonmin-enlarge < wind['lon']) & (wind['lon'] < lonmax+enlarge))[0]
                rows = np.where((latmin-enlarge < wind['lat']) & (wind['lat'] < latmax+enlarge))[0]
                focused_wind = {'wu': wind['wu'][:,rows,:][:,:,cols], 'wv': wind['wv'][:,rows,:][:,:,cols],
                                'lon': wind['lon'][cols],'lat': wind['lat'][rows],'wtime': wind['wtime'],'dataset': 'ERA5'}
            elif wind_product == 'COSMO':
                shape = wind['lat'].shape
                # Focus data to the area of interest in order to make the gridding lighter
                for r in range(1,shape[0]):
                    if np.min((latgrid.reshape((1,-1))-wind['lat'][r,:].reshape((-1,1)))**2+(longrid.reshape((1,-1))-wind['lon'][r,:].reshape((-1,1)))**2) < 1e-4:
                        row_start = r-1
                        break
                for r in range(shape[0]-3,0,-1):
                    if np.min((latgrid.reshape((1,-1))-wind['lat'][r,:].reshape((-1,1)))**2+(longrid.reshape((1,-1))-wind['lon'][r,:].reshape((-1,1)))**2) < 1e-4:
                        row_end = r+2
                        break
                for c in range(1,shape[1]):
                    if np.min((latgrid.reshape((1,-1))-wind['lat'][row_start:row_end,c].reshape((-1,1)))**2+(longrid.reshape((1,-1))-wind['lon'][row_start:row_end,c].reshape((-1,1)))**2) < 1e-4:
                        col_start = c-1
                        break
                for c in range(shape[1]-3,0,-1):
                    if np.min((latgrid.reshape((1,-1))-wind['lat'][row_start:row_end,c].reshape((-1,1)))**2+(longrid.reshape((1,-1))-wind['lon'][row_start:row_end,c].reshape((-1,1)))**2) < 1e-4:
                        col_end = c+2
                        break
                focused_wind = {'wu': wind['wu'][:,row_start:row_end,col_start:col_end],'wv': wind['wv'][:,row_start:row_end,col_start:col_end],
                                'lon': wind['lon'][row_start:row_end,col_start:col_end],'lat': wind['lat'][row_start:row_end,col_start:col_end],
                                'wtime': wind['wtime'],'dataset': 'COSMO'}
            print(f"Computing divergence map for source '{source_name}'...")
            start_time = time.process_time()
            if source_name not in maps_xr.keys():
                divergence_xr = local_divergence_map(data_path,focused_wind,start_date,end_date,latgrid,longrid,latgrid_km,longrid_km,
                                                     sats=sats,
                                                     co2_noise_scenario=co2_noise_scenario,
                                                     co2_cloud_threshold=co2_cloud_threshold,
                                                     no2_noise_scenario=no2_noise_scenario,
                                                     no2_cloud_threshold=no2_cloud_threshold,
                                                     trace_gases=trace_gases)
                maps_xr[source_name] = divergence_xr
            else:
                divergence_xr = maps_xr[source_name]
                
            end_time = time.process_time()
            print(f'...done in {end_time-start_time} seconds')
            # Simple time stamp reading according to the naming of the SMARTCARB files 
            timestamps = list(map(lambda s: datetime.strptime(str(s)[:19],'%Y-%m-%dT%H:%M:%S'),divergence_xr['timestamps'].data))
            # Divergence map for trace gas
            div_map = divergence_xr[f'{gas}_div'].data
            # Create pixels x pixels sized kilometer grid around the source
            pixels = 4
            row_slice = slice((len(latgrid_km)-1)//2-pixels,(len(latgrid_km)+1)//2+pixels,1)
            col_slice = slice((len(longrid_km)-1)//2-pixels,(len(longrid_km)+1)//2+pixels,1)
            Y, X = np.meshgrid(latgrid_km[row_slice],longrid_km[col_slice],indexing='ij')
            
            start_time = time.process_time()
            print(f"Optimizing trace gas {gas} emissions for source '{source_name}' using least squares...")
            if gas == 'NO2':
                init = [900, 5, 6, 0.0, 0.0, 0.0, -0.5]
                bounds = np.array([(0,2000), (1,9), (1,9), (-10,10), (-10,10), (-1,1), (-1,1)])
                scales = [1e3,1e0,1e0,1e0,1e0,1e0,1e-1]
            elif gas == 'CO2':
                init = [9e5, 7, 7, 0.0, 0.0, 0.0, -500]
                bounds = np.array([(0,2e6), (1,9), (1,9), (-10,10), (-10,10), (-1,1), (-1000,1000)])
                scales = [1e5,1e0,1e0,1e0,1e0,1e0,1e2]
            # Divergence near the source
            div_local = div_map[row_slice, col_slice]
            residual = lambda params: model_residual(params,(X,Y),div_local)
            optimized = least_squares(residual, init, x_scale=scales, bounds=bounds.T, method='trf')
            opt = optimized.x
            J = optimized.jac
            cov = np.linalg.inv(J.T @ J)
            end_time = time.process_time()
            print(f'...done in {end_time-start_time} seconds')
            
            # Optimization using adaptive metropolis algorithm
            use_AM = AM_options is not None
            if use_AM:
                print(f"Optimizing trace gas emissions for source '{source_name}' using adaptive Metropolis...")
                start_time = time.process_time()
                warmup, samples = AM_options['warmup'], AM_options['samples']
                sampling = AM_MC(residual,opt,samples,warmup,C0=np.diag(0.01*opt**2),
                                 bounds=bounds,init_fit=False,progress=False)
                opt, cov = sampling['mu'], sampling['cov']
                end_time = time.process_time()
                print(f'...done in {end_time-start_time} seconds')
            
            # NO2 emissions estimation and standard deviation
            emis_est, emis_err = opt[0], np.sqrt(cov[0,0])
            bg_mean = opt[-1]
            div_modeled = np.full((len(latgrid_km), len(longrid_km)), bg_mean)
            div_modeled[row_slice,col_slice] += peak_model(np.append(opt[:-1],0), (X,Y))
            #latgrid = latgrid[row_slice,col_slice]
            #longrid = longrid[row_slice,col_slice]
            emis_true = emissions[source_name.replace(' ', '_')].loc[timestamps].mean()
            
            out[gas][source_name]['lon'] = source_coord[0]
            out[gas][source_name]['lat'] = source_coord[1]
            if gas == 'NO2':
                out[gas][source_name]['estimation'] = emis_est*3600*24*365.24219/10**9
                out[gas][source_name]['error'] = emis_err*3600*24*365.24219/10**9
                out[gas][source_name]['truth'] = 1000 * emis_true
            elif gas == 'CO2':
                out[gas][source_name]['estimation'] = emis_est*3600*24*365.24219/10**12
                out[gas][source_name]['error'] = emis_err*3600*24*365.24219/10**12
                out[gas][source_name]['truth'] = emis_true
            out[gas][source_name]['bg_mean'] = bg_mean
            out[gas][source_name]['x_correction'] = opt[3]
            out[gas][source_name]['y_correction'] = opt[4]
            out[gas][source_name]['images'] = divergence_xr['images'].data
            out[gas][source_name]['modeled_div'] = div_modeled
            out[gas][source_name]['computed_div'] = div_map
            out[gas][source_name]['latgrid'] = latgrid
            out[gas][source_name]['longrid'] = longrid
            out[gas][source_name]['lat_km'] = lat_km
            out[gas][source_name]['lon_km'] = lon_km
            out[gas][source_name]['km_reso'] = grid_reso
    
    # Construct output xarray object
    data_vars = {}
    dims = ['trace_gases', 'sources']
    coords = {'trace_gases': trace_gases, 'sources': source_names}
    
    lon_locations =  np.array([[out[gas][source_name]['lon'] for source_name in source_names] for gas in trace_gases])
    attrs = {'Description': 'Longitude coordinate of the source', 'Units': '°'}
    data_vars['lon_loc'] = xr.DataArray(data=lon_locations, dims=dims, coords=coords, attrs=attrs)
    lat_locations =  np.array([[out[gas][source_name]['lat'] for source_name in source_names] for gas in trace_gases])
    attrs = {'Description': 'Latitude coordinate of the source', 'Units': '°'}
    data_vars['lat_loc'] = xr.DataArray(data=lat_locations, dims=dims, coords=coords, attrs=attrs)
    x_corrections =  np.array([[out[gas][source_name]['x_correction'] for source_name in source_names] for gas in trace_gases])
    attrs = {'Description': 'Correction of the center pixel in x-direction', 'Units': 'km'}
    data_vars['x_correction'] = xr.DataArray(data=x_corrections, dims=dims, coords=coords, attrs=attrs)
    y_corrections = np.array([[out[gas][source_name]['y_correction'] for source_name in source_names] for gas in trace_gases])
    attrs = {'Description': 'Correction of the center pixel in y-direction', 'Units': 'km'}
    data_vars['y_correction'] = xr.DataArray(data=y_corrections, dims=dims, coords=coords, attrs=attrs)
    
    units = {'CO2': 'Mt/y', 'NO2': 'kt/y'}
    estimations = np.array([[out[gas][source_name]['estimation'] for source_name in source_names] for gas in trace_gases])
    attrs = {'Description': 'Estimated emissions', 'Units': units, 'Missing Value': np.nan}
    data_vars['estimations'] = xr.DataArray(data=estimations, dims=dims, coords=coords, attrs=attrs)
    error = np.array([[out[gas][source_name]['error'] for source_name in source_names] for gas in trace_gases])
    attrs = {'Description': 'Estimation errors', 'Units': units, 'Missing Value': np.nan}
    data_vars['error'] = xr.DataArray(data=error, dims=dims, coords=coords, attrs=attrs)
    truth = np.array([[out[gas][source_name]['truth'] for source_name in source_names] for gas in trace_gases])
    attrs = {'Description': 'True emissions', 'Units': units, 'Missing Value': np.nan}
    data_vars['truth'] = xr.DataArray(data=truth, dims=dims, coords=coords, attrs=attrs)
    background = np.array([[out[gas][source_name]['bg_mean'] for source_name in source_names] for gas in trace_gases])
    attrs = {'Description': 'Background mean', 'Units': 'kg/m2/s', 'Missing Value': np.nan}
    data_vars['background'] = xr.DataArray(data=background, dims=dims, coords=coords, attrs=attrs)
    
    lats = np.stack([np.stack([out[gas][source_name]['latgrid'] for source_name in source_names]) for gas in trace_gases])
    lons = np.stack([np.stack([out[gas][source_name]['longrid'] for source_name in source_names]) for gas in trace_gases])
    dims = ['trace_gases','sources','lat','lon']
    data_vars['longrid'] = xr.DataArray(data=lons, dims=dims, attrs={'Decription': 'Longitude grid', 'Units': '°'})
    data_vars['latgrid'] = xr.DataArray(data=lats, dims=dims, attrs={'Decription': 'Latitude grid', 'Units': '°'})
    model = np.stack([np.stack([out[gas][source_name]['modeled_div'] for source_name in source_names]) for gas in trace_gases])
    attrs = {'Description': 'Divergence map', 'Units': 'kg/m2/s', 'Missing Value': np.nan}
    data_vars['modeled_divergence'] = xr.DataArray(data=model, dims=dims, attrs=attrs)
    computed = np.stack([np.stack([out[gas][source_name]['computed_div'] for source_name in source_names]) for gas in trace_gases])
    attrs = {'Description': 'Divergence map', 'Units': 'kg/m2/s', 'Missing Value': np.nan}
    data_vars['computed_divergence'] = xr.DataArray(data=computed, dims=dims, attrs=attrs)
    #dims = ['trace_gases','sources','images']
    estimation_end = time.process_time()
    print(f'Estimations done in {estimation_end-estimation_start} seconds')
    return xr.Dataset(data_vars=data_vars,attrs={'Desciption':'Emission estimations using divergence method'})

def estimate_emission_global(source_dict, data_path, wind_path,
                             center, lon_km=50, lat_km=50, grid_reso=5, UTC='11:00:00',
                             sats=['0000','0805','1610'], wind_product='ERA5',
                             co2_noise_scenario='low', co2_cloud_threshold=0.01,
                             no2_noise_scenario='low', no2_cloud_threshold=0.30,
                             start_date='2015-01-01', end_date='2015-12-24',
                             AM_options=None):
    """ 
    Estimates the NOx and CO2 emissions for one source over the chosen time series
    by computing the divergence map using the SMARTCARB data with the chosen satellites
    and fitting a peak fitting model.
    """
    start_date = datetime.strptime(start_date, '%Y-%m-%d') if isinstance(start_date, str) else start_date
    end_date = datetime.strptime(end_date, '%Y-%m-%d') if isinstance(end_date, str) else end_date
    print('Reading wind data...')
    wind = read_wind_data(wind_path, start_date, end_date, dataset=wind_product)
    print('...done')
    print('Reading emission data...')
    avg_emissions_CO2,avg_emissions_NOx = read_emission_data(start_date,end_date,UTC,trace_gases=('CO2','NO2'))
    print('...done')

    # Date ordinals from start_date to end_date
    ordinals = np.arange(start_date.toordinal(),end_date.toordinal()+1,1,dtype=int)
    dates = [datetime.fromordinal(ordinal) for ordinal in ordinals]

    # Regular km grid of the region
    longrid, latgrid, longrid_km, latgrid_km = grids(center,lon_km,lat_km,grid_reso)
    latmin, latmax = np.min(latgrid), np.max(latgrid)
    lonmin, lonmax = np.min(longrid), np.max(longrid)
    # Focus wind field to the area of interest to save time during interpolation
    print('Focusing wind...')
    if wind_product == 'ERA5':
        enlarge = 0.5
        cols = np.where((lonmin-enlarge < wind['lon']) & (wind['lon'] < lonmax+enlarge))[0]
        rows = np.where((latmin-enlarge < wind['lat']) & (wind['lat'] < latmax+enlarge))[0]
        focused_wind = {'wu': wind['wu'][:,rows,:][:,:,cols], 'wv': wind['wv'][:,rows,:][:,:,cols],
                        'lon': wind['lon'][cols],'lat': wind['lat'][rows],'wtime': wind['wtime'],'dataset': 'ERA5'}
    elif wind_product == 'COSMO':
        shape = wind['lat'].shape
        for r in range(1,shape[0]):
            if np.min((latgrid.reshape((1,-1))-wind['lat'][r,:].reshape((-1,1)))**2+(longrid.reshape((1,-1))-wind['lon'][r,:].reshape((-1,1)))**2) < 1e-4:
                row_start = r-1
                break
        for r in range(shape[0]-3,0,-1):
            if np.min((latgrid.reshape((1,-1))-wind['lat'][r,:].reshape((-1,1)))**2+(longrid.reshape((1,-1))-wind['lon'][r,:].reshape((-1,1)))**2) < 1e-4:
                row_end = r+2
                break
        for c in range(1,shape[1]):
            if np.min((latgrid.reshape((1,-1))-wind['lat'][row_start:row_end,c].reshape((-1,1)))**2+(longrid.reshape((1,-1))-wind['lon'][row_start:row_end,c].reshape((-1,1)))**2) < 1e-4:
                col_start = c-1
                break
        for c in range(shape[1]-3,0,-1):
            if  np.min((latgrid.reshape((1,-1))-wind['lat'][row_start:row_end,c].reshape((-1,1)))**2+(longrid.reshape((1,-1))-wind['lon'][row_start:row_end,c].reshape((-1,1)))**2) < 1e-4:
                col_end = c+2
                break
        focused_wind = {'wu': wind['wu'][:,row_start:row_end,col_start:col_end],'wv': wind['wv'][:,row_start:row_end,col_start:col_end],
                        'lon': wind['lon'][row_start:row_end,col_start:col_end],'lat': wind['lat'][row_start:row_end,col_start:col_end],
                        'wtime': wind['wtime'],'dataset': 'COSMO'}
    print('...done')
    print("Computing divergence map...")
    divergence_xr = global_divergence_map(data_path,focused_wind,dates,latgrid,longrid,latgrid_km,longrid_km,
                                          sats=sats,
                                          co2_noise_scenario=co2_noise_scenario,
                                          co2_cloud_threshold=co2_cloud_threshold,
                                          no2_noise_scenario=no2_noise_scenario,
                                          no2_cloud_threshold=no2_cloud_threshold)
    print('...done')
    div_NOx = divergence_xr['NOx divergence'].data
    div_CO2 = divergence_xr['CO2 divergence'].data
    init_NOx = [900, 5, 6, None, None, 0.0, -0.5, 0.2]
    bounds_NOx = [(0,2000), (1,9), (1,9), None, None, (-1,1), (-1,1), (0,1)]
    init_CO2 = [9e5, 7, 7, None, None, 0.0, -500, 200]
    bounds_CO2 = [(0,2e6), (1,9), (1,9), None, None, (-1,1), (-1000,1000), (0,1500)]
    crop_sizes = {'Belchatow': 3,
                  'Berlin': 4,
                  'Boxberg': 2,
                  'Chvaletice': 3,
                  'Dolna Odra': 3,
                  'Hamburg': 3,
                  'Heyden': 3,
                  'Janschwalde': 3,
                  'Lippendorf': 3,
                  'Melnik': 3,
                  'Opole': 3,
                  'Patnow': 3,
                  'Pocerady': 3,
                  'Prague': 3,
                  'Prunerov': 3,
                  'Schkopau': 3,
                  'Schwarze Pumpe': 2,
                  'Staudinger': 3,
                  'Turow': 3}
    # Initialize out dictionary
    out = {}
    for source_name in source_dict.keys():
        crop_size = crop_sizes[source_name]
        x0,y0,ind_loc,X_crop,Y_crop = around_source(source_name,longrid,longrid_km,latgrid,latgrid_km,crop_size=crop_size)
        source_coord = (x0,y0)
        row_slice = slice(ind_loc[0]-crop_size,ind_loc[0]+crop_size+1,1)
        col_slice = slice(ind_loc[1]-crop_size,ind_loc[1]+crop_size+1,1)
        Y, X = np.meshgrid(latgrid_km[row_slice],longrid_km[col_slice],indexing='ij')
        
        init_NOx[3],init_NOx[4] = x0, y0
        init_CO2[3],init_CO2[4] = x0, y0
        bounds_CO2[3],bounds_CO2[4] = (x0-2*grid_reso,x0+2*grid_reso),(y0-2*grid_reso,y0+2*grid_reso)
        bounds_NOx[3],bounds_NOx[4] = (x0-2*grid_reso,x0+2*grid_reso),(y0-2*grid_reso,y0+2*grid_reso)
        # Initial values and options for optimization
        options = {'maxiter': 10000}
        method = 'Powell'
        
        print(f"Optimizing emissions for source '{source_name}'...")
        
        # Optimization first using scipy.optimize package
        div_NOx_modeled = div_NOx[row_slice, col_slice]
        res_NOx = minimize(peak_model_ss, x0=init_NOx, args=(X, Y, div_NOx_modeled),
                           bounds=bounds_NOx, options=options, method=method)
        opt_NOx = res_NOx.x
        cov_NOx = np.linalg.inv(hessian(peak_model_ss, opt_NOx, args=(X, Y, div_NOx_modeled)))
        
        div_CO2_modeled = div_CO2[row_slice, col_slice]
        res_CO2 = minimize(peak_model_ss,x0=init_CO2,args=(X_crop, Y_crop, div_CO2_modeled),
                           bounds=bounds_CO2, options=options, method=method)
        opt_CO2 = res_CO2.x
        cov_CO2 = np.linalg.inv(hessian(peak_model_ss,opt_CO2,args=(X_crop, Y_crop, div_CO2_modeled)))
        # Optimization using adaptive metropolis algorithm
        use_AM = AM_options is not None
        if use_AM:
            warmup, maxiter = AM_options['warmup'], AM_options['maxiter']
            opt_NOx, cov_NOx, chain_NOx = AM(peak_model_ss, opt_NOx, np.diag((0.1*np.abs(opt_NOx))**2),
                                             maxiter, warmup, args=(X, Y, div_NOx_modeled), bounds=bounds_NOx)
            opt_CO2, cov_CO2, chain_CO2 = AM(peak_model_ss, opt_CO2, np.diag((0.1*np.abs(opt_CO2))**2),
                                             maxiter, warmup, args=(X_crop, Y_crop, div_CO2_modeled), bounds=bounds_CO2)
        print("...done")
        emis_est_NOx, emis_err_NOx = opt_NOx[0], np.sqrt(cov_NOx[0,0])
        NOx_mean, NOx_std = opt_NOx[-2], opt_NOx[-1]
        #NOx_modeled = np.full((len(latgrid_km), len(longrid_km)), NOx_mean)
        #NOx_modeled[row_slice, col_slice] += peak_model(opt_NOx, (X, Y))

        CO2_mean, CO2_std = opt_CO2[-2], opt_CO2[-1]
        emis_est_CO2, emis_err_CO2 = opt_CO2[0], np.sqrt(cov_CO2[0,0])
        #CO2_modeled = np.full((len(latgrid_km), len(longrid_km)), CO2_mean)
        #CO2_modeled[row_slice, col_slice] += peak_model(opt_CO2, (X_crop, Y_crop))
        try:
            # True emission values
            true_NOx = avg_emissions_NOx[source_name.replace(' ', '_')]
            true_CO2 = avg_emissions_CO2[source_name.replace(' ', '_')]
        except KeyError:
            true_NOx = np.nan
            true_CO2 = np.nan
        out[source_name] = {'true loc (lon,lat)': source_coord,
                            'NOx est (kt/y)': 1000*emis_est_NOx*3600*24*365/10**12,
                            'NOx err (kt/y)': 1000*emis_err_NOx*3600*24*365/10**12,
                            'NOx true (kt/y)': 1000*true_NOx,
                            'NOx chain': chain_NOx if use_AM else np.empty(0),
                            'NOx bg mean (kg/m2/s)': NOx_mean,
                            'NOx bg std (kg/m2/s)': NOx_std,
                            'NOx opt loc (x,y)': (opt_NOx[3], opt_NOx[4]),
                            'CO2 est (Mt/y)': emis_est_CO2*3600*24*365/10**12,
                            'CO2 err (Mt/y)': emis_err_CO2*3600*24*365/10**12,
                            'CO2 true (Mt/y)': true_CO2,
                            'CO2 chain': chain_CO2 if use_AM else np.empty(0),
                            'CO2 bg mean (kg/m2/s)': CO2_mean,
                            'CO2 bg std (kg/m2/s)': CO2_std,
                            'CO2 opt loc (x,y)': (opt_CO2[3], opt_CO2[4]),
                            'images': divergence_xr['images'].data,
                            'true NOx div': div_NOx,
                            #'modeled NOx div': NOx_modeled,
                            #'modeled CO2 div': CO2_modeled,
                            'true CO2 div': div_CO2,
                            'longrid': longrid,
                            'latgrid': latgrid,
                            'km_reso': grid_reso}
    return out

def plot_divergence_maps(NOx_div, CO2_div, lon_lat_center, lon_km, lat_km, grid_reso,
                         sources=None, ms=5, fontsize='x-large', cbar_fraction=0.051):
    longrid, latgrid, _, _ = grids(lon_lat_center, lon_km+grid_reso/2, lat_km+grid_reso/2, grid_reso)
    request = cimgt.OSM()
    #request = cimgt.QuadtreeTiles()
    fig, (ax1, ax2) = plt.subplots(1, 2, sharex=True, sharey=True, figsize=(16, 10),
                                   subplot_kw=dict(projection=request.crs))
    # plot NOx divergence
    extent = [np.min(longrid)-0.2, np.max(longrid)+0.2,
              np.min(latgrid)-0.2, np.max(latgrid)+0.2]
    ax1.set_extent(extent)
    ax1.add_image(request, 6)

    if sources is not None:
        ddeq.vis.add_hot_spots(ax1, sources=sources, size=fontsize, ms=ms)
    c1 = ax1.pcolormesh(longrid, latgrid, NOx_div*1000,
                        shading='auto',
                        cmap=plt.cm.jet,
                        transform=ccrs.PlateCarree(),
                        alpha=0.3,
                        vmin=-1000,
                        vmax=2000)

    ax1.set_title("NO$_x$ divergence x 1000")
    ax1.set_xlabel('longitude (°)')
    ax1.set_ylabel('latitude (°)')
    cbar1 = plt.colorbar(mappable=c1, ax=ax1, fraction=cbar_fraction)
    cbar1.ax.set_title('kg/m$^2$/s')

    # plot CO2 divergence
    ax2.set_extent(extent)
    ax2.add_image(request, 6)
    if sources is not None:
        ddeq.vis.add_hot_spots(ax2, sources=sources, size=fontsize, ms=ms)
    c2 = ax2.pcolormesh(longrid, latgrid, CO2_div,
                        shading='auto',
                        cmap=plt.cm.jet,
                        transform=ccrs.PlateCarree(),
                        alpha=0.3,
                        vmin=-1000,
                        vmax=2000)
    ax2.set_title("CO$_2$ divergence")
    ax2.set_xlabel('longitude (°)')
    ax2.set_ylabel('latitude (°)')
    cbar2 = plt.colorbar(mappable=c2, ax=ax2, fraction=cbar_fraction)
    cbar2.ax.set_title('kg/m$^2$/s')
    return fig

def grids(center, lon_km, lat_km, grid_reso):
    """
    Generate the km and degree grids corresponding resolution grid_reso in km.
    """
    longrid_km = np.arange(-lon_km, lon_km+grid_reso, grid_reso)
    latgrid_km = np.arange(-lat_km, lat_km+grid_reso, grid_reso)[::-1]
    p = LatLon(center[1],center[0])
    lat_arr = [p.destination(d*1000,0).latlon2(ndigits=5).lat for d in latgrid_km]
    rows, cols = len(latgrid_km), len(longrid_km)
    lat = np.repeat(np.reshape(lat_arr, (-1, 1)), cols, axis=1)
    lon = np.full([rows, cols], np.nan)
    for r in range(rows):
        p = LatLon(lat_arr[r],center[0])
        lon[r,:] = [p.destination(d*1000, 90).latlon2(ndigits=5).lon for d in longrid_km]
    return lon, lat, longrid_km, latgrid_km

def read_wind_data(path, start_date, end_date, dataset='ERA5'):
    dates = [(start_date + timedelta(days=n)) for n in range((end_date - start_date).days + 1)]
    if dataset == 'ERA5':
        #PATH_ABS = '/home/janne-nurmela/project/coco2/jupyter/WP4/ERA5/'
        files = [f for f in os.listdir(path) if f.startswith('ERA5-gnfra') and f.endswith('.nc') and f[20:24] == '1100'
                 and int(str(start_date.date()).replace('-', '')) <= int(f[11:19]) and int(f[11:19]) <= int(str(end_date.date()).replace('-', ''))]
        files.sort(key=lambda s: int(s[11:19]))
        datasets = [Dataset(path + f) for f in files]
        wlon, wlat = datasets[0]['longitude'][:].data, datasets[0]['latitude'][:].data
        # Days since 1900-01-01
        since = datetime(1900, 1, 1).toordinal()
        days = np.array([d.toordinal() - since for d in dates], dtype=int)
        wu = np.stack(tuple(data['U_GNFR_A'][:, :, :].data.squeeze()
                      for data in datasets), axis=0)
        wv = np.stack(tuple(data['V_GNFR_A'][:, :, :].data.squeeze()
                      for data in datasets), axis=0)
        wind = {'wu': wu, 'wv': wv, 'lon': wlon,
                'lat': wlat, 'wtime': days, 'dataset': 'ERA5'}
    elif dataset == 'COSMO':
        #PATH_ABS = '/home/janne-nurmela/project/coco2/fileshare/WP4/SMARTCARB/wind_fields/'
        UTC_time = '11:00:00'
        #levels = [37]
        files = sorted(os.listdir(path))
        datetimes = [datetime(int(d.year), int(d.month), int(d.day), int(UTC_time[:2]), int(UTC_time[3:5]), int(UTC_time[6:]))
                     for d in dates]
        datasets = [Dataset(path + 'SMARTCARB_winds_' + str(d.date()).replace('-', '') + UTC_time[:2] + '.nc') for d in dates]
        #wu = np.stack(tuple(data['U'][:,levels,:,:].mean(axis=1).squeeze() for data in files),axis=0)
        #wv = np.stack(tuple(data['V'][:,levels,:,:].mean(axis=1).squeeze() for data in files),axis=0)
        wu = np.stack(tuple(data['U_GNFR_A'][:, :, :].squeeze()
                      for data in datasets), axis=0)
        wv = np.stack(tuple(data['V_GNFR_A'][:, :, :].squeeze()
                      for data in datasets), axis=0)
        #wu_10m = np.stack(tuple(data['U_10M'][:,:,:].squeeze() for data in files),axis=0)
        #wv_10m = np.stack(tuple(data['V_10M'][:,:,:].squeeze() for data in files),axis=0)
        wlat, wlon = datasets[0]['lat'][:, :], datasets[0]['lon'][:, :]
        # Days since 1900-01-01
        since = datetime(1900, 1, 1).toordinal()
        days = np.array([d.toordinal() - since for d in dates], dtype=int)
        wind = {'wu': wu[:,::-1,:], 'wv': wv[:,::-1,:], 'lon': wlon[::-1,:], 'lat': wlat[::-1,:],
                'wtime': days, 'datetimes': datetimes, 'dataset': 'COSMO'}
    return wind

def read_emission_data(start_date,end_date,UTC,trace_gases=('CO2','NO2')):
    if 'CO2' in trace_gases:
        emissions_CO2 = pd.read_csv(ddeq_path+'/ddeq/data/ts_CO2_emissions.dat', delim_whitespace=True)
        emissions_CO2 = emissions_CO2[emissions_CO2['UTC']== UTC].reset_index(drop=True)
        emissions_CO2.index = emissions_CO2['date'].apply(lambda s: datetime.strptime(s, '%Y-%m-%d'))
        emissions_CO2.drop(columns=['date'], inplace=True)
        start = emissions_CO2.index.searchsorted(start_date)
        end = emissions_CO2.index.searchsorted(end_date)
        avg_emissions_CO2 = emissions_CO2.iloc[start:end+1].mean(numeric_only=True)
    if 'NO2' in trace_gases:
        emissions_NO2 = pd.read_csv(ddeq_path+'/ddeq/data/ts_NO2_emissions.dat', delim_whitespace=True)
        emissions_NO2 = emissions_NO2[emissions_NO2['UTC'] == UTC].reset_index(drop=True)
        emissions_NO2.index = emissions_NO2['date'].apply(lambda s: datetime.strptime(s, '%Y-%m-%d'))
        emissions_NO2.drop(columns=['date'], inplace=True)
        start = emissions_NO2.index.searchsorted(start_date)
        end = emissions_NO2.index.searchsorted(end_date)
        #Average over time series
        avg_emissions_NO2 = emissions_NO2.iloc[start:end+1].mean(numeric_only=True)
    if 'CO2' in trace_gases and 'NO2' in trace_gases:
        return avg_emissions_CO2, avg_emissions_NO2
    elif 'CO2' in trace_gases and 'NO2' not in trace_gases:
        return avg_emissions_CO2
    elif 'CO2' not in trace_gases and 'NO2' in trace_gases:
        return avg_emissions_NO2
    else:
        return None

def hessian(fun, x, args=()):
    '''
    Hessian approximation using central difference approximation.

    Parameters
    ----------
    fun : callable
        Function from R^n to R
    x : tuple
        The point in which the Hessian is approximated.
    args : tuple, optional
        Other arguments passed to fun.

    Returns
    -------
    H : 2d-array
        Hessian matrix of fun at x.
    '''
    x = np.array(x)
    # Optimal step sizes for finite difference
    scales = np.array([10**np.ceil(np.log10(abs(xi))) if xi != 0.0 else 1 for xi in x])
    h = 1e-6*scales
    dim = len(x)
    # Unit vectors to each spatial direction
    e = np.eye(dim)
    H = np.zeros(shape=(dim, dim))
    # Diagonal elements
    for d in range(dim):
        H[d, d] = (-fun(x+2*h*e[:, d], *args)+16*fun(x+h*e[:, d], *args)-30*fun(x, *args)
                   + 16*fun(x-h*e[:, d], *args)-fun(x-2*h*e[:, d], *args)) / (12*h[d]**2)
    # Off-diagonal elements
    for r in range(dim-1):
        for c in range(r+1, dim):
            H[r, c] = (fun(x+h*e[:, r]+h*e[:, c], *args) - fun(x+h*e[:, r]-h*e[:, c], *args)
                       - fun(x-h*e[:, r]+h*e[:, c], *args) + fun(x-h*e[:, r]-h*e[:, c], *args)) / (4*h[r]*h[c])
            H[c, r] = H[r, c]
    return H