#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 15 10:48:43 2022

@author: nurmelaj
"""

import numpy as np
from scipy.spatial import distance
from pygeodesy.ellipsoidalVincenty import LatLon
from scipy.interpolate import RectSphereBivariateSpline,SmoothSphereBivariateSpline
import glob
import fnmatch
import xarray as xr
import sys
from datetime import datetime
import ddeq

def local_divergence_map(data_path,wind,start_date,end_date,latgrid,longrid,latgrid_km,longrid_km,
                         sats = ['0000','0805','1610'],
                         co2_noise_scenario='low', co2_cloud_threshold=0.01,
                         no2_noise_scenario='low', no2_cloud_threshold=0.30,
                         trace_gases=['CO2']):
    dates = [datetime.fromordinal(ordinal) for ordinal in range(start_date.toordinal(),end_date.toordinal()+1)]
    l, m, n = len(dates), len(latgrid_km), len(longrid_km)
    #Validate wind dictionary keys
    wu_key = validate_keys(wind,['wu','vx','u'])
    wv_key = validate_keys(wind,['wv','vy','v'])
    lon_key = validate_keys(wind,['lon','wlon','longitude'])
    lat_key = validate_keys(wind,['lat','wlat','latitude'])
    
    NO2_trace = 'NO2' in trace_gases
    #Note: tracing gas NO2 actually corresponds NOx = NO + NO2 and NO2 / NOx ≈ 0.957
    CO2_trace = 'CO2' in trace_gases
    
    if NO2_trace:
        NO2in = np.full([l, m, n], np.nan)
        Fx_NO2in = np.full([l, m, n], np.nan)
        Fy_NO2in = np.full([l, m, n], np.nan)
    if CO2_trace:
        CO2in = np.full([l, m, n], np.nan)
        Fx_CO2in = np.full([l, m, n], np.nan)
        Fy_CO2in = np.full([l, m, n], np.nan)
    
    images = []
    timestamps = []
    for i in range(l):
        #All files from date
        filenames = glob.glob(data_path + '/*' + dates[i].strftime('%Y%m%d') + '*')
        #Take files with chosen overpasses
        files = [files[0] for files in (fnmatch.filter(filenames, '*l' + sat + '*') for sat in sats) if len(files) != 0]
        if len(files) == 0:
            print('SMARTCARB data not found!')
            continue
        num = np.zeros([m,n])
        if NO2_trace:
            total_NO2 = np.zeros([m,n])
            total_Fx_NO2 = np.zeros([m,n])
            total_Fy_NO2 = np.zeros([m,n])
        if CO2_trace :
            total_CO2 = np.zeros([m,n])
            total_Fx_CO2 = np.zeros([m,n])
            total_Fy_CO2 = np.zeros([m,n])
        #print(str(dates[i].date()))
        for f in range(len(files)):
            file = ddeq.smartcarb.read_level2(files[f],
                                              co2_noise_scenario=co2_noise_scenario, 
                                              co2_cloud_threshold=co2_cloud_threshold, 
                                              co2_scaling=1.0,
                                              no2_noise_scenario=no2_noise_scenario,  
                                              no2_cloud_threshold=no2_cloud_threshold, 
                                              no2_scaling=1.0,
                                              add_background=True)
            lon = file.lon.data.flatten()
            lat = file.lat.data.flatten()
            # Surface pressure
            psurf = file.psurf.data.flatten()
            # NO2 data
            if NO2_trace:
                NO2 = file.NO2.data.flatten()
                # NO2 units from molec/cm^2 to molec/m^2
                NO2 = 10000 * NO2
            if CO2_trace:
                # CO2 data and smoothing
                CO2 = file.CO2.data
                CO2 = ddeq.misc.normalized_convolution(CO2, np.ones(shape=(5,5))/25).flatten()
            
            # Indices of the data values which are inside the chosen grid
            #enlarge = 0.5 if wind['dataset'] == 'ERA5' else 0.0
            inside_ind = indices_inside_grid(lon,lat,longrid,latgrid,enlarge=0.1)
            if len(inside_ind) == 0:
                continue
            lon, lat = lon[inside_ind], lat[inside_ind]
            if NO2_trace:
                NO2 = NO2[inside_ind]
            if CO2_trace:
                CO2 = CO2[inside_ind]
            psurf = psurf[inside_ind]
            #Check if there is any actual data pixels for any of the trace gases
            if NO2_trace and not CO2_trace:
                if np.all(np.isnan(NO2)):
                    continue
            elif not NO2_trace and CO2_trace:
                if np.all(np.isnan(CO2)):
                    continue
            else:
                if np.all(np.isnan(NO2)) and np.all(np.isnan(CO2)):
                    continue
            #Append the file name/image
            images.append(files[f])
            # Append timestamp to list if it does not already exist in the list
            timestamp = datetime.strptime(files[f].split('_')[3],'%Y%m%d%H')
            if len(timestamps) == 0:
                timestamps.append(timestamp)
            elif timestamp != timestamps[-1]:
                timestamps.append(timestamp)
            
            # XCO2 anomaly inside the chosen grid, excluding points with distance less than 10 km to the grid center
            center = LatLon(latgrid[(m-1)//2,(n-1)//2],longrid[(m-1)//2,(n-1)//2])
            dists_km = np.array([LatLon(coord[0],coord[1]).distanceTo(center)/1000 for coord in zip(lat,lon)])
            
            if CO2_trace:
                # Remove CO2 background
                CO2 = CO2 - np.nanmedian(CO2[np.where(dists_km >= 10)])
                # CO2 units from ppm to molec/m^2
                CO2 = CO2*6.0221409e23/9.807/0.02896*psurf/1e6  
            
            #Read wind data
            wu, wv = wind[wu_key][i,:,:], wind[wv_key][i,:,:]
            wlon, wlat = wind[lon_key], wind[lat_key]
            #Interpolate wind in spherical coordinates
            if wind['dataset'] == 'ERA5':
                wu_interpolator = RectSphereBivariateSpline(np.deg2rad(90-wlat),np.deg2rad(wlon),wu)
                wv_interpolator = RectSphereBivariateSpline(np.deg2rad(90-wlat),np.deg2rad(wlon),wv)
            elif wind['dataset'] == 'COSMO':
                wu_flat, wv_flat = wu.flatten(), wv.flatten()
                wu_weights, wv_weights = np.ones(len(wu_flat))/np.sqrt(len(wu_flat)), np.ones(len(wv_flat))/np.sqrt(len(wv_flat))
                wu_interpolator = SmoothSphereBivariateSpline(np.deg2rad(90-wlat.flatten()),np.deg2rad(wlon.flatten()),wu_flat,w=wu_weights,s=1e-2*len(wu_flat))
                wv_interpolator = SmoothSphereBivariateSpline(np.deg2rad(90-wlat.flatten()),np.deg2rad(wlon.flatten()),wv_flat,w=wv_weights,s=1e-2*len(wv_flat))
            wu_interp = wu_interpolator.ev(np.deg2rad(90-lat),np.deg2rad(lon))
            wv_interp = wv_interpolator.ev(np.deg2rad(90-lat),np.deg2rad(lon))
            
            # Fluxes
            if NO2_trace:
                Fx_NO2 = NO2*wu_interp
                Fy_NO2 = NO2*wv_interp
            if CO2_trace:
                Fx_CO2 = CO2*wu_interp
                Fy_CO2 = CO2*wv_interp
            
            # gridding, undersampling
            for r in range(m):
                if r == 0:
                    lat_diff_north = np.mean(latgrid[0,:] - latgrid[1,:])
                    lat_diff_south = latgrid[r,0]-latgrid[r+1,0]
                elif r == m-1:
                    lat_diff_north = latgrid[r-1,0]-latgrid[r,0]
                    lat_diff_south = np.mean(latgrid[-2,:] - latgrid[-1,:])
                else:
                    lat_diff_north = latgrid[r-1,0]-latgrid[r,0]
                    lat_diff_south = latgrid[r,0]-latgrid[r+1,0]
                ind_lat = np.where((latgrid[r,0]-lat_diff_south/2 <= lat) & (lat <= latgrid[r,0]+lat_diff_north/2))[0]
                if not ind_lat.any():
                    continue
                for c in range(n):
                    if c == 0:
                        lon_diff_east = longrid[r,c+1]-longrid[r,c]
                        lon_diff_west = longrid[r,1] - longrid[r,0]
                    elif c == n-1:
                        lon_diff_east = longrid[r,-1]-longrid[r,-2]
                        lon_diff_west = longrid[r,c]-longrid[r,c-1]
                    else:
                        lon_diff_east = longrid[r,c+1]-longrid[r,c]
                        lon_diff_west = longrid[r,c]-longrid[r,c-1]
                    ind_lon = np.where((longrid[r,c]-lon_diff_west/2 <= lon[ind_lat]) & (lon[ind_lat] <= (longrid[r,c]+lon_diff_east/2)))[0]
                    if not ind_lon.any():
                        continue
                    num[r,c] += 1
                    if NO2_trace:
                        total_NO2[r,c] += np.nanmean(NO2[ind_lat][ind_lon])
                        total_Fx_NO2[r,c] += np.nanmean(Fx_NO2[ind_lat][ind_lon])
                        total_Fy_NO2[r,c] += np.nanmean(Fy_NO2[ind_lat][ind_lon])
                    if CO2_trace:
                        total_CO2[r,c] += np.nanmean(CO2[ind_lat][ind_lon])
                        total_Fx_CO2[r,c] += np.nanmean(Fx_CO2[ind_lat][ind_lon])
                        total_Fy_CO2[r,c] += np.nanmean(Fy_CO2[ind_lat][ind_lon])
        # Set zeros to nan
        num[np.where(num == 0.0)] = np.nan
        # Pixel averaging
        if NO2_trace:
            NO2in[i,:,:] = total_NO2/num
            Fx_NO2in[i,:,:] = total_Fx_NO2/num
            Fy_NO2in[i,:,:] = total_Fy_NO2/num
        if CO2_trace:
            CO2in[i,:,:] = total_CO2/num
            Fx_CO2in[i,:,:] = total_Fx_CO2/num
            Fy_CO2in[i,:,:] = total_Fy_CO2/num
    dims = ['lat', 'lon']
    if NO2_trace:
        # NO2 fluxes with NOx correction, about 95.71 % of NOx is NO2
        Fx_NO2 = np.nanmean(Fx_NO2in, axis=0)*0.9571
        Fy_NO2 = np.nanmean(Fy_NO2in, axis=0)*0.9571
        # Divergence of NO2 in g/m²/s or Mg/km²/s
        div_NO2 = 46.0055/6.0221409e23*divergence([Fx_NO2, Fy_NO2],[longrid_km*1e3, latgrid_km*1e3])
        # Conversion from Mg/km²/s to g/km²/s
        div_NO2 = div_NO2*1e6
        # NOx sinks
        sinks_NO2 = 46.0055/6.0221409e23*0.9571*np.nanmean(NO2in, axis=0)/(4*3600)
        sinks_NO2 = sinks_NO2*1e6
        div_NO2_xr = xr.DataArray(div_NO2,dims=dims,attrs={'long name':'NOx_divergence','units':'Mt/y'})
        #sinks_NOx_xr = xr.DataArray(sinks_NOx,dims=dims,attrs={'long name': 'NOx_sinks','units': 'Mt/y'})
        #emissions_NOx_xr = xr.DataArray((div_NOx+sinks_NOx),dims=dims,attrs={'long name': 'NOx_emissions','units': 'Mt/y'})
    if CO2_trace:
        # CO2 fluxes
        Fx_CO2 = np.nanmean(Fx_CO2in, axis=0)
        Fy_CO2 = np.nanmean(Fy_CO2in, axis=0)
        # Divergence of CO2 in g/m²/s or Mg/km²/s
        div_CO2 = 44.01/6.0221409e23*divergence([Fx_CO2, Fy_CO2],[longrid_km*1e3,latgrid_km*1e3])
        # Conversion from Mg/km²/s to g/km²/s
        div_CO2 = div_CO2*1e6
        div_CO2_xr = xr.DataArray(div_CO2,dims=dims,attrs={'long name':'CO2_divergence','units':'Mt/y'})
    #Save to xarray dataset
    data_vars = {'timestamps': xr.DataArray(timestamps,dims=['n_times']),
                 'longitude': xr.DataArray(longrid, dims=dims),
                 'latitude': xr.DataArray(latgrid, dims=dims),
                 'images': xr.DataArray(images,dims=['n_images'])}
    if NO2_trace:
        data_vars['NO2_div'] = div_NO2_xr
    if CO2_trace:
        data_vars['CO2_div'] = div_CO2_xr
    attrs = {'Description': 'Divergence map around the chosen source'}
    divergence_xr = xr.Dataset(data_vars=data_vars,attrs=attrs)
    return divergence_xr

def global_divergence_map(data_path,wind,dates,latgrid,longrid,latgrid_km,longrid_km,
                          sats = ['0000','0805','1610'],
                          co2_noise_scenario='low', co2_cloud_threshold=0.01,
                          no2_noise_scenario='low', no2_cloud_threshold=0.30,
                          start_date='2015-01-01',end_date='2015-12-24'):
    l, m, n = len(dates), len(latgrid_km), len(longrid_km)
    #Validate wind dictionary keys
    wu_key = validate_keys(wind,['wu','vx','u'])
    wv_key = validate_keys(wind,['wv','vy','v'])
    lon_key = validate_keys(wind,['lon','wlon','longitude'])
    lat_key = validate_keys(wind,['lat','wlat','latitude'])
        
    NO2in = np.full([l, m, n], np.nan)
    CO2in = np.full([l, m, n], np.nan)
    Fx_NO2in = np.full([l, m, n], np.nan)
    Fy_NO2in = np.full([l, m, n], np.nan)
    Fx_CO2in = np.full([l, m, n], np.nan)
    Fy_CO2in = np.full([l, m, n], np.nan)
    
    images = []
    for i in range(l):
        #print('Date:',dates[i].date())
        filenames = glob.glob(data_path + '/*' + dates[i].strftime('%Y%m%d') + '*')
        files = [files[0] for files in (fnmatch.filter(filenames, '*l' + sat + '*') for sat in sats) if len(files) != 0]
        if len(files) == 0:
            print('SMARTCARB data not found!')
            continue
        num = np.zeros([m,n])
        total_NO2 = np.zeros([m,n])
        total_CO2 = np.zeros([m,n])
        total_Fx_NO2 = np.zeros([m,n])
        total_Fy_NO2 = np.zeros([m,n])
        total_Fx_CO2 = np.zeros([m,n])
        total_Fy_CO2 = np.zeros([m,n])
        for f in range(len(files)):
            file = ddeq.smartcarb.read_level2(files[f],
                        co2_noise_scenario=co2_noise_scenario, 
                        co2_cloud_threshold=co2_cloud_threshold, 
                        co2_scaling=1.0,
                        no2_noise_scenario=no2_noise_scenario, 
                        no2_cloud_threshold=no2_cloud_threshold, 
                        no2_scaling=1.0,
                        add_background=True)
            lon = file.lon.data.flatten()
            lat = file.lat.data.flatten()
            
            # Surface pressure
            psurf = file.psurf.data.flatten()
            # NO2 data
            NO2 = file.NO2.data.flatten()
            # CO2 data and smoothing
            CO2 = file.CO2.data
            CO2 = ddeq.misc.normalized_convolution(CO2, np.ones(shape=(5,5))/25).flatten()
            
            # Indices of the data values which are inside the chosen grid
            inside_ind = indices_inside_grid(lon,lat,longrid,latgrid,enlarge=0.05)
            if len(inside_ind) == 0:
                #print(f"No data inside the chosen grid for sats='{sats[f]}'")
                continue
            images.append(files[f])
            lon, lat = lon[inside_ind], lat[inside_ind]
            NO2, CO2, psurf = NO2[inside_ind], CO2[inside_ind], psurf[inside_ind]
            # change NO2 units from molec/cm^2 to molec/m^2
            NO2 = 10000 * NO2
            # compute XCO2 anomaly inside the chosen grid
            CO2 = CO2 - np.nanmedian(CO2)
            # change CO2 units from ppm to molec/m^2
            CO2 = CO2*6.0221409e23/9.807/0.02896*psurf/1e6
            
            #Read wind data
            wu, wv = wind[wu_key][i,:,:], wind[wv_key][i,:,:]
            wlon, wlat = wind[lon_key], wind[lat_key]
            
            if wind['dataset'] == 'ERA5':
                wu_interpolator = RectSphereBivariateSpline(np.deg2rad(90-wlat),np.deg2rad(wlon),wu)
                wv_interpolator = RectSphereBivariateSpline(np.deg2rad(90-wlat),np.deg2rad(wlon),wv)
            elif wind['dataset'] == 'COSMO':
                wu_flat, wv_flat = wu.flatten(), wv.flatten()
                wu_weights, wv_weights = np.ones(len(wu_flat))/np.sqrt(len(wu_flat)), np.ones(len(wv_flat))/np.sqrt(len(wv_flat))
                wu_interpolator = SmoothSphereBivariateSpline(np.deg2rad(90-wlat.flatten()),np.deg2rad(wlon.flatten()),wu_flat,w=wu_weights,s=1e-2*len(wu_flat))
                wv_interpolator = SmoothSphereBivariateSpline(np.deg2rad(90-wlat.flatten()),np.deg2rad(wlon.flatten()),wv_flat,w=wv_weights,s=1e-2*len(wv_flat))
            wu_interp = wu_interpolator.ev(np.deg2rad(90-lat),np.deg2rad(lon))
            wv_interp = wv_interpolator.ev(np.deg2rad(90-lat),np.deg2rad(lon))
            #Fluxes
            Fx_NO2 = NO2*wu_interp
            Fy_NO2 = NO2*wv_interp
            Fx_CO2 = CO2*wu_interp
            Fy_CO2 = CO2*wv_interp
            
            # gridding, undersampling
            for r in range(1,m-1):
                ind_lat = np.where((latgrid[r,0]-(latgrid[r,0]-latgrid[r+1,0])/2 <= lat) & (lat <= latgrid[r,0]+(latgrid[r-1,0]-latgrid[r,0])/2))[0]
                if not ind_lat.any():
                    continue
                for c in range(1,n-1):
                    ind_lon = np.where((longrid[r,c]-(longrid[r,c]-longrid[r,c-1])/2 <= lon[ind_lat]) & (lon[ind_lat] <= (longrid[r,c]+(longrid[r,c+1]-longrid[r,c])/2)))[0]
                    if not ind_lon.any():
                        continue
                    num[r,c] += 1
                    total_NO2[r,c] += np.nanmean(NO2[ind_lat][ind_lon])
                    total_CO2[r,c] += np.nanmean(CO2[ind_lat][ind_lon])
                    total_Fx_NO2[r,c] += np.nanmean(Fx_NO2[ind_lat][ind_lon])
                    total_Fy_NO2[r,c] += np.nanmean(Fy_NO2[ind_lat][ind_lon])
                    total_Fx_CO2[r,c] += np.nanmean(Fx_CO2[ind_lat][ind_lon])
                    total_Fy_CO2[r,c] += np.nanmean(Fy_CO2[ind_lat][ind_lon])
        # Set zeros to nan
        num[np.where(num == 0.0)] = np.nan
        # Pixel averaging
        NO2in[i,:,:] = total_NO2/num
        CO2in[i,:,:] = total_CO2/num
        Fx_NO2in[i,:,:] = total_Fx_NO2/num
        Fy_NO2in[i,:,:] = total_Fy_NO2/num
        Fx_CO2in[i,:,:] = total_Fx_CO2/num
        Fy_CO2in[i,:,:] = total_Fy_CO2/num
    # NOx
    Fx_NO2 = np.nanmean(Fx_NO2in, axis=0)*0.9571
    Fy_NO2 = np.nanmean(Fy_NO2in, axis=0)*0.9571
    # Divergence of NO2 in kg/m^2/s
    div_NOx = 46.0055/(6.0221409e23*1000)*divergence([Fx_NO2, Fy_NO2],[longrid_km*1e3, latgrid_km*1e3])
    #index = np.unravel_index(np.argmax(div_NOx),div_NOx.shape)
    # Divergence of NO2 in kg/pixel/s
    #div_NOx = (grid_res*1000)**2*div_NOx
    div_NOx = div_NOx*1e9
    sinks_NOx = 46.0055/(6.0221409e23*1000)*0.9571*np.nanmean(NO2in, axis=0)/(4*3600)
    #sinks_NOx = sinks_NOx*1e9
    # CO2
    Fx_CO2 = np.nanmean(Fx_CO2in, axis=0)
    Fy_CO2 = np.nanmean(Fy_CO2in, axis=0)
    # Divergence of CO2 in kg/m^2/s
    div_CO2 = 44.01/(6.0221409e23*1000)*divergence([Fx_CO2, Fy_CO2],[longrid_km*1e3,latgrid_km*1e3])
    # Divergence of NO2 in kg/pixel/s
    #div_CO2 = (grid_res*1000)**2*div_CO2
    div_CO2 = div_CO2*1e9
    
    # Put data into xarray dataset
    dims = ['lat', 'lon']
    div_NOx_xr = xr.DataArray(div_NOx, dims=dims,attrs={'long name': 'NOx_divergence','units': 'Mt/y'})
    sinks_NOx_xr = xr.DataArray(sinks_NOx, dims=dims,attrs={'long name': 'NOx_sinks','units': 'Mt/y'})
    emissions_NOx_xr = xr.DataArray((div_NOx+sinks_NOx),dims=dims,
                                    attrs={'long name': 'NOx_emissions','units': 'Mt/y'})
    div_CO2_xr = xr.DataArray(div_CO2,dims=dims,attrs={'long name': 'CO2_divergence','units': 'Mt/y'})
    
    images_xr = xr.DataArray(images,dims=['n_images'],attrs={'long name':'images'})
    divergence_xr = xr.Dataset(data_vars={'longitude': xr.DataArray(longrid, dims=dims),
                                          'latitude': xr.DataArray(latgrid, dims=dims),
                                          'images': images_xr,
                                          'NOx divergence': div_NOx_xr,
                                          'NOx sinks': sinks_NOx_xr,
                                          'NOx emissions': emissions_NOx_xr,
                                          'CO2 divergence': div_CO2_xr})
    return divergence_xr

def around_source(source, longrid, longrid_km, latgrid, latgrid_km, crop_size=3):
    """
    This function croppes an area around the source, and returns a grid in km
    
    Input:  source = name of emission source
            crop_size = number of pixels to each direction from the source
    Output: x0, y0 = coordinates of the source wrt the new grid in km
            ind_loc = position of source in the grid
            X_crop, Y_crop = grid in km of area around the source
    """
    
    # list of point sources
    sources = ddeq.misc.read_point_sources()
    sources_coord = {sources.source.data[i]: [sources.variables['lon_o'].data[i],sources.variables['lat_o'].data[i]]
                     for i in range(len(sources.source.data))}
    coord = sources_coord[source]
    
    # check if the source is in our aoi
    if not ((coord[0] > np.min(longrid)) & (coord[0] < np.max(longrid)) 
            & (coord[1] > np.min(latgrid)) & (coord[1] < np.max(latgrid))):
        msg = "The source is not inside the area of interest"
        sys.exit(msg)
        
    # Crop an area around the selected source
    names, lon_coors, lat_coors = sources['source'].data, sources['lon_o'].data, sources['lat_o'].data
    sources_location = {names[i]: np.asarray([lon_coors[i], lat_coors[i]]) for i in range(len(names))}

    coords = np.asarray([[longrid.ravel()[i],latgrid.ravel()[i]] for i in range(np.prod(longrid.shape))])

    ind = distance.cdist([sources_location[source]], coords).argmin()
    ind_loc = np.unravel_index(ind,(len(latgrid_km),len(longrid_km)))

    x0 = longrid_km[ind_loc[1]]
    y0 = latgrid_km[ind_loc[0]]
    
    Y, X = np.meshgrid(latgrid_km,longrid_km,indexing='ij')
    #Slice array around the source
    X_crop = X[ind_loc[0]-crop_size:ind_loc[0]+crop_size+1,ind_loc[1]-crop_size:ind_loc[1]+crop_size+1]
    Y_crop = Y[ind_loc[0]-crop_size:ind_loc[0]+crop_size+1,ind_loc[1]-crop_size:ind_loc[1]+crop_size+1]
    return x0, y0, ind_loc, X_crop, Y_crop

def validate_keys(dictionary,valid_keys):
    for key in valid_keys + [None]:
        if key in dictionary.keys():
            break
    if key is not None:
        return key
    else:
        sys.exit(f'No valid keys in the dictionary for keys {valid_keys}')
    
def divergence(f, sp):
    """ 
    Computes divergence of vector field 
    f: array -> vector field components [Fx,Fy,Fz,...]
    sp: array -> spacing between points in respective directions 
    [spx,spy,spz,...]
    """
    d = len(f)
    grad = [np.gradient(f[i], sp[i], axis=d-1-i) for i in range(d)]
    return np.ufunc.reduce(np.add, grad)
    
def indices_inside_grid(data_x,data_y,grid_x,grid_y,enlarge=0.0):
    if grid_x.ndim == 1 and grid_y.ndim == 1:
        indices = np.where((np.min(grid_x)-enlarge <= data_x) & (data_x <= np.max(grid_x)+enlarge) &
                           (np.min(grid_y)-enlarge <= data_y) & (data_y <= np.max(grid_y)+enlarge))[0]
    elif grid_x.ndim == 2 and grid_y.ndim == 2:
        # Four lines limiting the grid
        top_line = lambda x : (grid_y[0,-1]-grid_y[0,0]) / (grid_x[0,-1]-grid_x[0,0]) * (x-grid_x[0,0]) + grid_y[0,0] + enlarge
        bottom_line = lambda x : (grid_y[-1,-1]-grid_y[-1,0]) / (grid_x[-1,-1]-grid_x[-1,0]) * (x-grid_x[-1,-1]) + grid_y[-1,-1] - enlarge
        left_line = lambda y : (grid_x[-1,0]-grid_x[0,0]) / (grid_y[-1,0]-grid_y[0,0]) * (y-grid_y[0,0]) + grid_x[0,0] - enlarge
        right_line = lambda y : (grid_x[-1,-1]-grid_x[0,-1]) / (grid_y[-1,-1]-grid_y[0,-1]) * (y-grid_y[0,-1]) + grid_x[0,-1] + enlarge
        indices = np.where((bottom_line(data_x) <= data_y) & (data_y <= top_line(data_x)) & 
                           (left_line(data_y) <= data_x) & (data_x <= right_line(data_y)))[0]
    return indices

def peak_model(theta, grids):
    """
    Model that we want to fit to each source
    
    theta = parameters that to be optimized
    grid = cropped grid around the source
    """
    # Peak integral corresponding emission in kg/s
    Q = theta[0]
    # Deviations x and y directions in km
    sigma_x, sigma_y = theta[1], theta[2]
    # Center of the peak
    x0, y0 = theta[3], theta[4]
    # Correlation between x and y dimensions
    corr = theta[5]
    # Background in kg/m²/s
    B = theta[6]
    # Kilometer grids
    X, Y = grids[0], grids[1]
    
    #mx = theta[7]
    #my = theta[8]
    #mx, my = 0, 0
    
    #G = A/(2*np.pi*sigma_x*sigma_y)*np.exp(-((X-x0)**2/(2*sigma_x**2)+(Y-y0)**2/(2*sigma_y**2)))+mx*(X-x0)+my*(Y-y0)+B
    # Normalization
    N = Q/(2*np.pi*sigma_x*sigma_y*np.sqrt(1-corr**2))
    G = N*np.exp(-((X-x0)**2/(2*sigma_x**2*(1-corr**2))+(Y-y0)**2/(2*sigma_y**2*(1-corr**2))-corr*(X-x0)*(Y-y0)/(sigma_x*sigma_y*(1-corr**2)))) + B
    #G += mx*(X-x0)+my*(Y-y0)+B
    return G

def model_residual(theta, grids, data):
    """ 
    Residual function.
    """
    G = peak_model(theta, grids)
    non_nan = ~np.isnan(data)
    data = data[non_nan]
    residual = (G[non_nan] - data).ravel()
    #sigma_obs = theta[-1]
    #ss = 0.5*np.sum(np.square(residual))/sigma_obs**2 + 0.5*len(residual)*np.log(2*np.pi) + len(residual)*np.log(sigma_obs)
    return residual